package build

import (
	"sync"

	"git.akyoto.dev/cli/q/src/build/asm"
	fail "git.akyoto.dev/cli/q/src/errors"
)

// compile waits for the scan to finish and compiles all functions.
func compile(functions <-chan *Function, errors <-chan error) (Result, error) {
	result := Result{
		Unused: map[string]*Function{},
	}

	for functions != nil || errors != nil {
		select {
		case err, ok := <-errors:
			if !ok {
				errors = nil
				continue
			}

			return result, err

		case function, ok := <-functions:
			if !ok {
				functions = nil
				continue
			}

			result.Unused[function.Name] = function
		}
	}

	compileFunctions(result.Unused)

	for _, function := range result.Unused {
		if function.Error != nil {
			return result, function.Error
		}
	}

	main, exists := result.Unused["main"]

	if !exists {
		return result, fail.MissingMainFunction
	}

	result.Used = make([]*Function, 0, len(result.Unused))
	result.markAlive(main)
	result.findAliveCode(main)

	return result, nil
}

// compileFunctions starts a goroutine for each function compilation and waits for completion.
func compileFunctions(functions map[string]*Function) {
	wg := sync.WaitGroup{}

	for _, function := range functions {
		wg.Add(1)

		go func() {
			defer wg.Done()
			function.Compile()
		}()
	}

	wg.Wait()
}

// findAliveCode recursively finds all the calls to external functions and marks them as required.
func (result *Result) findAliveCode(caller *Function) {
	for _, x := range caller.Assembler.Instructions {
		if x.Mnemonic != asm.CALL {
			continue
		}

		name := x.Data.(*asm.Label).Name
		callee, exists := result.Unused[name]

		if !exists {
			continue
		}

		result.markAlive(callee)
		result.findAliveCode(callee)
	}
}

// markAlive marks a function as being alive.
func (result *Result) markAlive(f *Function) {
	result.Used = append(result.Used, f)
	result.InstructionCount += len(f.Assembler.Instructions)
	delete(result.Unused, f.Name)
}
